//! Super Simple Task Scheduler. Basically a basic version of CRON. Probably not very good.

use std::{
    sync::{Arc, RwLock},
    thread::{self, JoinHandle},
};

use chrono::{DateTime, Duration, Utc};
pub use cron::Schedule;

/// Very Simple Task Scheduler. Probably not super smart.
#[derive(Debug)]
pub struct Scheduler {
    threads:       Vec<JoinHandle<()>>,
    shutdown:      Arc<RwLock<bool>>,
    poll_interval: Duration,
}

impl Scheduler {
    /// Create a new [Scheduler] with a `59` second `poll_interval`. `poll_interval` represents the
    /// time between checks. Smaller times
    /// result in more accurate times, but potentially higher CPU usage.
    ///
    /// Use [Self::with_poll_interval] to set this interval.
    pub fn new() -> Self {
        Self {
            threads:       vec![],
            shutdown:      Arc::new(RwLock::new(false)),
            poll_interval: Duration::seconds(59),
        }
    }

    /// Create a new [Scheduler]. `poll_interval` represents the time between checks. Smaller times
    /// result in more accurate times, but potentially higher CPU usage.
    ///
    /// Note: If the `poll_interval` is too high, we may miss schedules. These will however be run
    /// in bulk on the next poll.
    pub fn with_poll_interval(poll_interval: Duration) -> Self {
        Self {
            threads: vec![],
            shutdown: Arc::new(RwLock::new(false)),
            poll_interval,
        }
    }

    /// Schedule a new task, this will be run on a new thread and will loop infinitely
    ///
    /// Each task callback gets passed a reference to `shutdown`. Setting this to true will cause
    /// all the threads to complete at their next iteration
    pub fn schedule<T>(&mut self, schedule: Schedule, func: T)
    where
        T: Fn(Arc<RwLock<bool>>) + Send + Sync + 'static,
    {
        let shutdown = self.shutdown.clone();

        let interval = self.poll_interval.clone().to_std().unwrap_or_default();

        self.threads.push(thread::spawn(move || {
            let mut last_run: DateTime<Utc> = Utc::now();

            while !*shutdown.read().unwrap() {
                thread::sleep(interval);

                let now = Utc::now();

                let mut run_count = 0;

                for cron in schedule.after(&last_run) {
                    if cron < now {
                        run_count += 1;
                    } else {
                        break;
                    }
                }

                for _ in 0..run_count {
                    func(shutdown.clone());

                    last_run = Utc::now();
                }
            }
        }));
    }
}

impl Default for Scheduler {
    fn default() -> Self {
        Self::new()
    }
}

impl Drop for Scheduler {
    fn drop(&mut self) {
        *self.shutdown.write().unwrap() = true;
    }
}
